@@ -269,7 +269,8 @@ for my $file (qw/bucardo Bucardo.pm/) {
 ## Prepare some test values for easy use
 ## The secondary names are for other databases, e.g. MySQL
 our %val;
-for (1..30) {
+my $xvalmax = 30;
+for (1..$xvalmax) {
     $val{SMALLINT}{$_} = $_;
     $val{INT}{$_} = 1234567+$_;
     $val{BIGINT}{$_} = 7777777777 + $_;
@@ -285,6 +286,10 @@ for (1..30) {
 }
 
 
+sub diag {
+    Test::More::diag(@_);
+}
+
 
 sub new {
 
@@ -306,7 +311,7 @@ sub new {
     $self->{file} = (caller)[1];
 
     ## Bail on first error? Default is ENV, then false.
-    $self->{bail} = exists $arg->{bail} ? $arg->{bail} : $ENV{BUCARDO_TESTBAIL} || 0;
+    $bail_on_error = exists $arg->{bail} ? $arg->{bail} : $ENV{BUCARDO_TESTBAIL} || 0;
 
     ## Name of the test schema
     $self->{schema} = 'bucardo_schema';
@@ -364,13 +369,13 @@ sub debug {
 
 sub empty_cluster {
 
-    ## Empty out a cluster's database
+    ## Empty out a cluster's databases
     ## Creates the cluster and 'bucardo_test' database as needed
     ## For existing databases, removes all known schemas
     ## Always recreates the public schema
-    ## Arguments:
+    ## Arguments: one
     ## 1. Name of the cluster
-    ## Returns: database handle to the 'bucardo_test' database
+    ## Returns: arrayref of database handles to the 'bucardo_test*' databases
 
     my $self = shift;
     my $clustername = shift or die;
@@ -381,10 +386,13 @@ sub empty_cluster {
     ## Start it up if needed
     $self->start_cluster($clustername);
 
-    ## Get a handle to the postgres database
-    my $dbh = $self->connect_database($clustername, 'postgres');
+    my $alldbh;
 
-    if (database_exists($dbh, $dbname)) {
+    ## Get a handle to the postgres database
+    my $masterdbh = $self->connect_database($clustername, 'postgres');
+
+    my $dbh;
+    if (database_exists($masterdbh, $dbname)) {
         $dbh = $self->connect_database($clustername, $dbname);
         ## Remove any of our known schemas
         my @slist;
@@ -398,13 +406,14 @@ sub empty_cluster {
         $dbh->commit();
     }
     else {
-        local $dbh->{AutoCommit} = 1;
+        local $masterdbh->{AutoCommit} = 1;
         debug(qq{Creating database $dbname});
-        $dbh->do("CREATE DATABASE $dbname");
-        $dbh->disconnect();
+        $masterdbh->do("CREATE DATABASE $dbname");
         $dbh = $self->connect_database($clustername, $dbname);
     }
 
+    $masterdbh->disconnect();
+
     return $dbh;
 
 } ## end of empty_cluster
@@ -421,8 +430,9 @@ sub create_cluster {
     my $self = shift;
     my $clustername = shift or die;
 
+    my $line = (caller)[2];
     my $info = $pgver{$clustername}
-        or die qq{No such cluster as "$clustername"\n};
+        or die qq{No such cluster as "$clustername" (called from line $line)\n};
 
     my $dirname = $info->{dirname};
 
@@ -448,12 +458,13 @@ sub create_cluster {
     }
 
     ## Make some minor adjustments
+    my $connections = $clustername eq 'A' ? 150 : 75;
     my $file = "$dirname/postgresql.conf";
     open my $fh, '>>', $file or die qq{Could not open "$file": $!\n};
     printf {$fh} "
 
 port                       = %d
-max_connections            = 20
+max_connections            = $connections
 random_page_cost           = 2.5
 log_statement              = 'all'
 log_min_duration_statement = 0
@@ -505,9 +516,6 @@ sub start_cluster {
         close $fh or die qq{Could not close "$pidfile": $!\n};
         ## An active process should respond to a "ping kill"
         $count = kill 0 => $pid;
-        #warn "GOT A count of $count for $pid!\n";
-        #my $count2 = kill 1 => $pid;
-        #warn "GOT A count of $count2 for $pid kill 1!\n";
         return if 1 == $count;
         ## If no response, remove the pidfile ourselves and go on
         debug(qq{Server seems to have died, removing file "$pidfile"});
@@ -590,6 +598,9 @@ sub connect_database {
     my $clustername = shift or die;
     my $ldbname = shift || $dbname;
 
+    ## This may be one of the "extra" databases. In which case the true cluster must be revealed:
+    $clustername =~ s/\d+$//;
+
     ## Create and start the cluster as needed
     $self->start_cluster($clustername);
 
@@ -682,16 +693,38 @@ sub drop_schema {
 sub repopulate_cluster {
 
     ## Make sure a cluster is empty, then add in the sample data
-    ## Arguments:
+    ## Arguments: two
     ## 1. Name of the cluster
+    ## 2. Optional - number of additional databases to create
     ## Returns: database handle to the 'bucardo_test' database
 
     my $self = shift;
     my $clustername = shift or die;
+    my $extradbs = shift || 0;
+
+    Test::More::note("Recreating cluster $clustername");
 
     my $dbh = $self->empty_cluster($clustername);
+    $self->add_test_schema($dbh, $clustername);
 
-    $self->add_test_schema($dbh);
+    ## Now recreate all the extra databases via templating
+    for my $number (1..$extradbs) {
+        my $dbname2 = "$dbname$number";
+        local $dbh->{AutoCommit} = 1;
+        if (database_exists($dbh, $dbname2)) {
+            ## First, kill other sessions!
+            my $odbh = $self->connect_database($clustername, $dbname2);
+            eval {
+                $SQL = 'SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = ? AND pid <> pg_backend_pid()';
+                $sth = $odbh->prepare($SQL);
+                $odbh->execute($dbname2);
+                $odbh->commit();
+            };
+            $odbh->disconnect();
+            $dbh->do("DROP DATABASE $dbname2");
+        }
+        $dbh->do("CREATE DATABASE $dbname2 TEMPLATE $dbname");
+    }
 
     ## Store our names away
     $gdbh{$clustername} = $dbh;
@@ -704,12 +737,14 @@ sub repopulate_cluster {
 sub add_test_schema {
 
     ## Add an empty test schema to a database
-    ## Arguments:
+    ## Arguments: two
     ## 1. database handle (usually to 'bucardo_test')
+    ## 2. Cluster name
     ## Returns: nothing
 
     my $self = shift;
     my $dbh = shift or die;
+    my $clustername = shift or die;
 
     my ($tcount,$scount,$fcount) = (0,0,0);
 
@@ -732,7 +767,6 @@ sub add_test_schema {
         debug(q{Creating language plpgsql});
         $dbh->do('CREATE LANGUAGE plpgsql');
     }
-    $dbh->commit() if ! $dbh->{AutoCommit};
 
     ## Create supporting functions as needed
     if (!function_exists($dbh => 'trigger_test')) {
@@ -839,15 +873,16 @@ ALTER TABLE bucardo_fkey1
 
         ## Does the sequence already exist? If so, drop it.
         if (table_exists($dbh => $seq)) {
-            $dbh->do("DROP SEQUENCE $seq");
+            $dbh->do(qq{DROP SEQUENCE "$seq"});
         }
 
-        $SQL = qq{CREATE SEQUENCE $seq};
+        $SQL = qq{CREATE SEQUENCE "$seq"};
         $dbh->do($SQL);
         $scount++;
     }
 
-    debug("Test objects created. Tables: $tcount  Sequences: $scount  Functions: $fcount");
+    debug("Test objects created for $clustername. Tables: $tcount  Sequences: $scount  Functions: $fcount");
+#    diag("Test objects created for $clustername. Tables: $tcount  Sequences: $scount  Functions: $fcount");
 
     $dbh->commit() if ! $dbh->{AutoCommit};
 
@@ -1054,14 +1089,15 @@ sub restart_bucardo {
     $dbh->do('LISTEN bucardo_nosyncs');
     $dbh->commit();
 
-    my $output = $self->ctl('start --exit-on-nosync testing');
+    my $output = $self->ctl('start --exit-on-nosync --quickstart testing');
 
-    my $bail = 30;
+    my $bail = 50;
     my $n;
   WAITFORIT: {
         if ($bail--<0) {
             $output =~ s/^/#     /gmx;
-            die "Bucardo did not start, but we waited! Start output:\n\n$output\n";
+            my $time = localtime;
+            die "Bucardo did not start, but we waited!\nTime: $time\nStart output:\n\n$output\n";
         }
         while ($n = $dbh->func('pg_notifies')) {
             my ($name, $pid, $payload) = @$n;
@@ -1104,6 +1140,8 @@ sub setup_bucardo {
     my $self = shift;
     my $clustername = shift or die;
 
+    Test::More::note('Installing Bucardo');
+
     $self->create_cluster($clustername);
     my $dbh = $self->connect_database($clustername, 'postgres');
     if (database_exists($dbh,'bucardo')) {
@@ -1195,7 +1233,7 @@ sub wait_for_notice {
     my $timeout = shift || $TIMEOUT_NOTICE;
     my $sleep = shift || $TIMEOUT_SLEEP;
     my $bail = shift;
-    $bail = 1 if !defined($bail);
+    $bail = 0 if !defined($bail);
     my $n;
     my %wait_for;
     for my $str (ref $text ? @{ $text } : $text) {
@@ -1377,9 +1415,19 @@ sub shutdown_cluster {
 
     my $pidfile = "$dirname/postmaster.pid";
     return if ! -e $pidfile;
+
+    Test::More::note("Stopping cluster $name");
     my @cmd = ($pg_ctl, '-D', $dirname, '-s', '-m', 'fast', 'stop');
     system(@cmd) == 0 or die "@cmd failed: $?\n";
 
+    ## Hang around until the PID file is gone
+    my $loops = 0;
+    {
+        sleep 0.2;
+        last if ! -e $pidfile;
+        redo;
+    }
+
     delete $gdbh{$name};
 
     return;
@@ -1565,6 +1613,11 @@ sub add_row_to_database {
 
     my ($self, $dbname, $xval, $commit) = @_;
 
+
+    if ($xval > $xvalmax) {
+        die "Too high of an ID: max is $xvalmax\n";
+    }
+
     $commit = 1 if ! defined $commit;
 
     my $dbh = $gdbh{$dbname} or die "No such database: $dbname";
@@ -1774,10 +1827,16 @@ sub check_for_row {
         if (defined $filter) {
             my $f = $filter;
             if ($f =~ s/^\!//) {
-                next if $table =~ /$f$/;
+                if ($table =~ /$f$/) {
+                    delete $tabletype{$table};
+                    next;
+                }
             }
             else {
-                next if $table !~ /$f$/;
+                if ($table !~ /$f$/) {
+                    delete $tabletype{$table};
+                    next;
+                }
             }
         }
         $maxtable = length $table if length $table > $maxtable;
@@ -1795,16 +1854,6 @@ sub check_for_row {
 
         for my $table (sort keys %tabletype) {
 
-            ## Allow skipping tables
-            if (defined $filter) {
-                my $f = $filter;
-                if ($f =~ s/^\!//) {
-                    next if $table =~ /$f$/;
-                }
-                else {
-                    next if $table !~ /$f$/;
-                }
-            }
             ## Handle odd pkeys
             my $pkey = $table =~ /test5/ ? q{"id space"} : 'id';
 
@@ -1855,7 +1904,7 @@ sub check_sequences_same {
 
     for my $seq (sort keys %sequences) {
 
-        $SQL = "SELECT * FROM $seq";
+        $SQL = qq{SELECT * FROM "$seq"};
 
         ## The first we come across will be the standard for the others
         my (%firstone, $firstdb);
@@ -1890,7 +1939,7 @@ sub check_sequences_same {
         if (@msg) {
             Test::More::fail("Sequence $seq NOT the same");
             for (@msg) {
-                Test::More::diag($_);
+                diag($_);
             }
         }
         else {
@@ -1920,8 +1969,8 @@ sub is_deeply {
     if ($bail_on_error and ++$total_errors => $bail_on_error) {
         my $line = (caller)[2];
         my $time = time;
-        Test::More::diag("GOT: ".Dumper $_[0]);
-        Test::More::diag("EXPECTED: ".Dumper $_[1]);
+        diag("GOT: ".Dumper $_[0]);
+        diag("EXPECTED: ".Dumper $_[1]);
         Test::More::BAIL_OUT "Stopping on a failed 'is_deeply' test from line $line. Time: $time";
     }
 } ## end of is_deeply
@@ -1958,7 +2007,7 @@ sub is($$;$) {
         my $one = ord(substr($_[0],$char,1));
         my $two = ord(substr($_[1],$char,1));
         if ($one != $two) {
-            Test::More::diag("First difference at character $char ($one vs $two) (line $line, char $lchar)");
+            diag("First difference at character $char ($one vs $two) (line $line, char $lchar)");
             last;
         }
         if (10 == $one) {
